# SPDX-FileCopyrightText: 2022 Contributors to the Power Grid Model project <dynamic.grid.calculation@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

"""
Helper functions to assert valid data. They basically call validate_input_data or validate_batch_data and raise a
ValidationException if the validation results in one or more errors.
"""
from typing import Dict, List, Optional, Union

from power_grid_model.data_types import BatchDataset, SingleDataset
from power_grid_model.enum import CalculationType
from power_grid_model.validation.errors import ValidationError
from power_grid_model.validation.utils import errors_to_string
from power_grid_model.validation.validation import validate_batch_data, validate_input_data


class ValidationException(ValueError):
    """
    An exception storing the name of the validated data, a list/dict of errors and a convenient conversion to string
    to display a summary of all the errors when printing the exception.
    """

    def __init__(self, errors: Union[List[ValidationError], Dict[int, List[ValidationError]]], name: str = "data"):
        super().__init__(f"Invalid {name}")
        self.errors = errors
        self.name = name

    def __str__(self):
        return errors_to_string(errors=self.errors, name=self.name)


def assert_valid_input_data(
    input_data: SingleDataset, calculation_type: Optional[CalculationType] = None, symmetric: bool = True
):
    """
    Validates the entire input dataset:

        1. Is the data structure correct? (checking data types and numpy array shapes)
        2. Are all required values provided? (checking NaNs)
        3. Are all ID's unique? (checking object identifiers across all components)
        4. Are the supplied values valid? (checking limits and other logic as described in "Graph Data Model")

    Args:
        input_data: A power-grid-model input dataset
        calculation_type: Supply a calculation method, to allow missing values for unused fields
        symmetric: A boolean to state whether input data will be used for a symmetric or asymmetric calculation

    Raises:
        KeyError, TypeError or ValueError if the data structure is invalid.
        ValidationException if the contents are invalid.
    """
    validation_errors = validate_input_data(
        input_data=input_data, calculation_type=calculation_type, symmetric=symmetric
    )
    if validation_errors:
        raise ValidationException(validation_errors, "input_data")


def assert_valid_batch_data(
    input_data: SingleDataset,
    update_data: BatchDataset,
    calculation_type: Optional[CalculationType] = None,
    symmetric: bool = True,
):
    """
    Ihe input dataset is validated:

        1. Is the data structure correct? (checking data types and numpy array shapes)
        2. Are all input data ID's unique? (checking object identifiers across all components)

    For each batch the update data is validated:
        3. Is the update data structure correct? (checking data types and numpy array shapes)
        4. Are all update ID's valid? (checking object identifiers across update and input data)

    Then (for each batch independently) the input dataset is updated with the batch's update data and validated:
        5. Are all required values provided? (checking NaNs)
        6. Are the supplied values valid? (checking limits and other logic as described in "Graph Data Model")

    Args:
        input_data: a power-grid-model input dataset
        update_data: a power-grid-model update dataset (one or more batches)
        calculation_type: Supply a calculation method, to allow missing values for unused fields
        symmetric: A boolean to state whether input data will be used for a symmetric or asymmetric calculation

    Raises:
        KeyError, TypeError or ValueError if the data structure is invalid.
        ValidationException if the contents are invalid.
    """
    validation_errors = validate_batch_data(
        input_data=input_data, update_data=update_data, calculation_type=calculation_type, symmetric=symmetric
    )
    if validation_errors:
        raise ValidationException(validation_errors, "update_data")



# SPDX-FileCopyrightText: 2022 Contributors to the Power Grid Model project <dynamic.grid.calculation@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

import json
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pytest

from power_grid_model.data_types import Dataset, SingleDataset
from power_grid_model.utils import export_json_data, import_json_data

BASE_PATH = Path(__file__).parent.parent
DATA_PATH = BASE_PATH / "data"
OUPUT_PATH = BASE_PATH / "output"
EXPORT_OUTPUT = ("POWER_GRID_MODEL_VALIDATION_TEST_EXPORT" in os.environ) and (
    os.environ["POWER_GRID_MODEL_VALIDATION_TEST_EXPORT"] == "ON"
)


def pytest_cases(get_batch_cases: bool = False, data_dir: Optional[str] = None, test_cases: Optional[List[str]] = None):
    if data_dir is not None:
        relevant_calculations = [data_dir]
    else:
        # default both calculation type
        relevant_calculations = ["power_flow", "state_estimation"]
    batch_suffix = "_batch" if get_batch_cases else ""
    for calculation_type in relevant_calculations:
        # list of all cases, directories in validation datasets
        calculation_type_dir = DATA_PATH / calculation_type
        test_cases_paths = {
            str(item.relative_to(DATA_PATH)).replace("\\", "/"): item
            for item in calculation_type_dir.glob("**/")
            if (item.is_dir() and (item / "params.json").is_file())
        }
        if test_cases is not None:
            test_cases_paths = {key: value for key, value in test_cases_paths.items() if key in test_cases}
        for case_name, case_dir in test_cases_paths.items():
            with open(case_dir / "params.json") as f:
                params = json.load(f)
            # retrieve calculation method, can be a string or list of strings
            calculation_methods = params["calculation_method"]
            if not isinstance(calculation_methods, list):
                calculation_methods = [calculation_methods]
            # loop for sym or asym scenario
            for sym in [True, False]:
                output_prefix = "sym_output" if sym else "asym_output"
                for calculation_method in calculation_methods:
                    # only generate a case if sym or asym output exists
                    if (case_dir / f"{output_prefix}{batch_suffix}.json").exists():
                        # Build a recognizable case ID
                        case_id = case_name
                        case_id += "-sym" if sym else "-asym"
                        case_id += "-" + calculation_method
                        if get_batch_cases:
                            case_id += "-batch"
                        pytest_param = [
                            case_id,
                            case_dir,
                            sym,
                            calculation_type,
                            calculation_method,
                            params["rtol"],
                            params["atol"],
                        ]
                        kwargs = {}
                        if "fail" in params:
                            kwargs["marks"] = pytest.mark.xfail(reason=params["fail"], raises=AssertionError)
                        yield pytest.param(*pytest_param, **kwargs, id=case_id)


def bool_params(true_id: str, false_id: Optional[str] = None, **kwargs):
    if false_id is None:
        false_id = f"not-{true_id}"
    yield pytest.param(False, **kwargs, id=false_id)
    yield pytest.param(True, **kwargs, id=true_id)


def dict_params(params: Dict[Any, str], **kwargs):
    for value, param_id in params.items():
        yield pytest.param(value, **kwargs, id=param_id)


def import_case_data(data_path: Path, sym: bool):
    output_prefix = "sym_output" if sym else "asym_output"
    return_dict = {
        "input": import_json_data(data_path / "input.json", "input", ignore_extra=True),
    }
    # import output if relevant
    if (data_path / f"{output_prefix}.json").exists():
        return_dict["output"] = import_json_data(data_path / f"{output_prefix}.json", output_prefix, ignore_extra=True)
    # import update and output batch if relevant
    if (data_path / "update_batch.json").exists():
        return_dict["update_batch"] = import_json_data(data_path / "update_batch.json", "update", ignore_extra=True)
        return_dict["output_batch"] = import_json_data(
            data_path / f"{output_prefix}_batch.json", output_prefix, ignore_extra=True
        )
    return return_dict


def save_json_data(json_file: str, data: Dataset):
    OUPUT_PATH.mkdir(parents=True, exist_ok=True)
    data_file = OUPUT_PATH / json_file
    data_file.parent.mkdir(parents=True, exist_ok=True)
    export_json_data(data_file, data)


def compare_result(actual: SingleDataset, expected: SingleDataset, rtol: float, atol: Union[float, Dict[str, float]]):
    for key, expected_data in expected.items():
        for col_name in expected_data.dtype.names:
            actual_col = actual[key][col_name]
            expected_col = expected_data[col_name]
            if expected_col.dtype == np.float64:
                expect_all_nan = np.all(np.isnan(expected_col))
            elif expected_col.dtype == np.int8:
                expect_all_nan = np.all(expected_col == np.iinfo("i1").min)
            elif expected_col.dtype == np.int32:
                expect_all_nan = np.all(expected_col == np.iinfo("i4").min)
            else:
                raise Exception(f"Unknown data type {expected_col.dtype}!")

            if not expect_all_nan:
                # permute expected_col if needed
                if expected_col.ndim == 1 and actual_col.ndim == 2:
                    if col_name == "u_angle":
                        # should be 120 and 240 degree lagging
                        expected_col = np.stack(
                            (expected_col, expected_col - 2.0 / 3.0 * np.pi, expected_col + 2.0 / 3.0 * np.pi), axis=-1
                        )
                        # reset angle to -pi to +pi range
                        expected_col = np.angle(np.exp(1j * expected_col))
                    else:
                        expected_col = expected_col.reshape(-1, 1)
                # for u_angle, the angle needs to be normalized
                # because any global angle shift is acceptable
                if col_name == "u_angle":
                    # set the u_angle of 0-th entry to zero
                    actual_col = actual_col - actual_col.ravel()[0]
                    expected_col = expected_col - expected_col.ravel()[0]
                    # convert to unity complex number, to avoid under and over flows
                    actual_col = np.exp(1j * actual_col)
                    expected_col = np.exp(1j * expected_col)
                    # convert to radians
                    actual_col = np.angle(actual_col)
                    expected_col = np.angle(expected_col)

                # Define the absolute tolerance based on the field name
                p = "default"
                if isinstance(atol, dict):
                    a = atol[p]
                    for pattern, tolerance in atol.items():
                        if re.fullmatch(pattern, col_name):
                            a = tolerance
                            p = f"r'{pattern}'"
                            break
                else:
                    a = atol

                if not np.allclose(actual_col, expected_col, rtol=rtol, atol=a):
                    msg = f"Not all values match for {key}.{col_name} (rtol={rtol}, atol={a}, pattern={p})"
                    print(f"\n{msg}")
                    print("Actual:     ", actual_col)
                    print("Expected:   ", expected_col)
                    print("Difference: ", actual_col - expected_col)
                    print("Matches:    ", np.isclose(actual_col, expected_col, rtol=rtol, atol=a))
                    raise AssertionError(msg)



# SPDX-FileCopyrightText: 2022 Contributors to the Power Grid Model project <dynamic.grid.calculation@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from copy import copy
from pathlib import Path

import numpy as np
import pytest

from power_grid_model import PowerGridModel
from power_grid_model.utils import convert_batch_dataset_to_batch_list

from .utils import EXPORT_OUTPUT, compare_result, import_case_data, pytest_cases, save_json_data

calculation_function_map = {
    "power_flow": PowerGridModel.calculate_power_flow,
    "state_estimation": PowerGridModel.calculate_state_estimation,
}


@pytest.mark.parametrize(
    ["case_id", "case_path", "sym", "calculation_type", "calculation_method", "rtol", "atol"],
    pytest_cases(get_batch_cases=False),
)
def test_single_validation(
    case_id: str, case_path: Path, sym: bool, calculation_type: str, calculation_method: str, rtol: float, atol: float
):
    # Initialization
    case_data = import_case_data(case_path, sym=sym)
    model = PowerGridModel(case_data["input"], system_frequency=50.0)

    # Normal calculation
    result = calculation_function_map[calculation_type](model, symmetric=sym, calculation_method=calculation_method)

    # Compare the results
    reference_result = case_data["output"]
    compare_result(result, reference_result, rtol, atol)

    # test get indexer
    for component_name, input_array in case_data["input"].items():
        ids_array = input_array["id"].copy()
        np.random.shuffle(ids_array)
        indexer_array = model.get_indexer(component_name, ids_array)
        # check
        assert np.all(input_array["id"][indexer_array] == ids_array)

    # export data if needed
    if EXPORT_OUTPUT:
        save_json_data(f"{case_id}.json", result)

    # test calculate with only node and source result
    result = calculation_function_map[calculation_type](
        model, symmetric=sym, calculation_method=calculation_method, output_component_types=["node", "source"]
    )
    assert set(result.keys()) == {"node", "source"}
    result = calculation_function_map[calculation_type](
        model, symmetric=sym, calculation_method=calculation_method, output_component_types={"node", "source"}
    )
    assert set(result.keys()) == {"node", "source"}


@pytest.mark.parametrize(
    ["case_id", "case_path", "sym", "calculation_type", "calculation_method", "rtol", "atol"],
    pytest_cases(get_batch_cases=True),
)
def test_batch_validation(
    case_id: str, case_path: Path, sym: bool, calculation_type: str, calculation_method: str, rtol: float, atol: float
):
    # Initialization
    case_data = import_case_data(case_path, sym=sym)
    model = PowerGridModel(case_data["input"], system_frequency=50.0)
    update_batch = case_data["update_batch"]
    update_list = convert_batch_dataset_to_batch_list(update_batch)
    reference_output_batch = case_data["output_batch"]
    reference_output_list = convert_batch_dataset_to_batch_list(reference_output_batch)

    # execute batch calculation by applying update method
    for update_data, reference_result in zip(update_list, reference_output_list):
        model_copy = copy(model)
        model_copy.update(update_data=update_data)
        result = calculation_function_map[calculation_type](
            model_copy, symmetric=sym, calculation_method=calculation_method
        )
        compare_result(result, reference_result, rtol, atol)

    # execute in batch one go
    for threading in [-1, 0, 1, 2]:
        result_batch = calculation_function_map[calculation_type](
            model, symmetric=sym, calculation_method=calculation_method, update_data=update_batch, threading=threading
        )
        result_list = convert_batch_dataset_to_batch_list(result_batch)
        for result, reference_result in zip(result_list, reference_output_list):
            compare_result(result, reference_result, rtol, atol)

    # export data if needed
    if EXPORT_OUTPUT:
        save_json_data(f"{case_id}.json", result_batch)



# SPDX-FileCopyrightText: 2022 Contributors to the Power Grid Model project <dynamic.grid.calculation@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0

from enum import IntEnum

import pytest

from power_grid_model.validation.errors import (
    ComparisonError,
    InvalidEnumValueError,
    InvalidIdError,
    MultiComponentValidationError,
    MultiFieldValidationError,
    SingleFieldValidationError,
    ValidationError,
)


def test_validation_error():
    error = ValidationError()
    assert str(error) == "An unknown validation error occurred."


def test_single_field_validation_error():
    error = SingleFieldValidationError(component="alpha", field="bravo", ids=[0, 1, 1, 2, 3, 5])
    assert error.component_str == "alpha"
    assert error.field_str == "'bravo'"
    assert str(error) == "Field 'bravo' is not valid for 6 alphas."


def test_multi_field_validation_error():
    error = MultiFieldValidationError(component="charlie", fields=["delta", "echo"], ids=[0, 1, 1, 2, 3, 5])
    assert error.component_str == "charlie"
    assert error.field_str == "'delta' and 'echo'"
    assert str(error) == "Combination of fields 'delta' and 'echo' is not valid for 6 charlies."

    with pytest.raises(ValueError, match="at least two fields"):
        MultiFieldValidationError(component="charlie", fields=["delta"], ids=[])


def test_multi_component_validation_error():
    error = MultiComponentValidationError(
        fields=[("foxtrot", "golf"), ("hotel", "india")],
        ids=[("foxtrot", 0), ("foxtrot", 1), ("foxtrot", 1), ("hotel", 2), ("hotel", 3), ("hotel", 5)],
    )
    assert error.component_str == "foxtrot/hotel"
    assert error.field_str == "foxtrot.golf and hotel.india"
    assert str(error) == "Fields foxtrot.golf and hotel.india are not valid for 6 foxtrots/hotels."

    with pytest.raises(ValueError, match="at least two fields"):
        MultiComponentValidationError(fields=[("foxtrot", "golf")], ids=[])

    with pytest.raises(ValueError, match="at least two components"):
        MultiComponentValidationError(fields=[("foxtrot", "golf"), ("foxtrot", "india")], ids=[])


def test_invalid_enum_value_error():
    class CustomType(IntEnum):
        pass

    error = InvalidEnumValueError(component="kilo", field="lima", ids=[1, 2, 3], enum=CustomType)
    assert error.component == "kilo"
    assert error.field == "lima"
    assert error.ids == [1, 2, 3]
    assert error.enum is CustomType
    assert str(error) == "Field 'lima' contains invalid CustomType values for 3 kilos."


def test_invalid_id_error():
    error = InvalidIdError(component="mike", field="november", ids=[1, 2, 3], ref_components=["oscar", "papa"])
    assert error.component == "mike"
    assert error.field == "november"
    assert error.ids == [1, 2, 3]
    assert error.ref_components == ["oscar", "papa"]
    assert str(error) == "Field 'november' does not contain a valid oscar/papa id for 3 mikes."


def test_comparison_error():
    error = ComparisonError(component="quebec", field="romeo", ids=[1, 2, 3], ref_value=0)
    assert error.component == "quebec"
    assert error.field == "romeo"
    assert error.ids == [1, 2, 3]
    assert error.ref_value == 0
    assert ComparisonError(component="", field="", ids=[], ref_value=123).ref_value_str == "123"
    assert ComparisonError(component="", field="", ids=[], ref_value=123.456).ref_value_str == "123.456"
    assert ComparisonError(component="", field="", ids=[], ref_value=0).ref_value_str == "zero"
    assert ComparisonError(component="", field="", ids=[], ref_value=0.0).ref_value_str == "zero"
    assert ComparisonError(component="", field="", ids=[], ref_value=1).ref_value_str == "one"
    assert ComparisonError(component="", field="", ids=[], ref_value=1.0).ref_value_str == "one"
    assert ComparisonError(component="", field="", ids=[], ref_value="sierra").ref_value_str == "sierra"


def test_error_context():
    error = ComparisonError(component="sierra", field="tango", ids=[1, 2, 3], ref_value=0)
    context = error.get_context()
    assert len(context) == 4
    assert context["component"] == "sierra"
    assert context["field"] == "'tango'"
    assert context["ids"] == [1, 2, 3]
    assert context["ref_value"] == "zero"

    context = error.get_context(id_lookup=["Uniform", "Victor"])  # Note that id 0 does not exist in the error ids
    assert context["ids"] == {1: "Victor", 2: None, 3: None}

    context = error.get_context(id_lookup={1: "Victor", 3: "Whiskey"})
    assert context["ids"] == {1: "Victor", 2: None, 3: "Whiskey"}


def test_error_context_tuple_ids():
    error = MultiComponentValidationError(fields=[("a", "x"), ("b", "y")], ids=[("a", 1), ("b", 2), ("a", 3)])
    context = error.get_context(id_lookup={1: "Victor", 3: "Whiskey"})
    assert context["ids"] == {("a", 1): "Victor", ("b", 2): None, ("a", 3): "Whiskey"}
